@@ -28,6 +28,7 @@
 package org.apache.hc.core5.testing.nio;
 
 import java.io.IOException;
+import java.net.InetAddress;
 import java.net.InetSocketAddress;
 import java.net.SocketAddress;
 import java.util.concurrent.ExecutionException;
@@ -39,8 +40,10 @@ import javax.net.ssl.SSLHandshakeException;
 import javax.net.ssl.SSLSession;
 
 import org.apache.hc.core5.concurrent.BasicFuture;
+import org.apache.hc.core5.concurrent.CompletingFutureContribution;
 import org.apache.hc.core5.concurrent.FutureCallback;
 import org.apache.hc.core5.concurrent.FutureContribution;
+import org.apache.hc.core5.function.Supplier;
 import org.apache.hc.core5.http.ContentType;
 import org.apache.hc.core5.http.HttpHost;
 import org.apache.hc.core5.http.HttpResponse;
@@ -52,7 +55,9 @@ import org.apache.hc.core5.http.impl.bootstrap.AsyncRequesterBootstrap;
 import org.apache.hc.core5.http.impl.bootstrap.AsyncServerBootstrap;
 import org.apache.hc.core5.http.impl.bootstrap.HttpAsyncRequester;
 import org.apache.hc.core5.http.impl.bootstrap.HttpAsyncServer;
+import org.apache.hc.core5.http.impl.routing.RequestRouter;
 import org.apache.hc.core5.http.nio.AsyncClientEndpoint;
+import org.apache.hc.core5.http.nio.AsyncServerExchangeHandler;
 import org.apache.hc.core5.http.nio.entity.StringAsyncEntityConsumer;
 import org.apache.hc.core5.http.nio.entity.StringAsyncEntityProducer;
 import org.apache.hc.core5.http.nio.ssl.BasicClientTlsStrategy;
@@ -62,7 +67,6 @@ import org.apache.hc.core5.http.nio.ssl.TlsSupport;
 import org.apache.hc.core5.http.nio.ssl.TlsUpgradeCapable;
 import org.apache.hc.core5.http.nio.support.BasicRequestProducer;
 import org.apache.hc.core5.http.nio.support.BasicResponseConsumer;
-import org.apache.hc.core5.http.protocol.UriPatternMatcher;
 import org.apache.hc.core5.http.ssl.TLS;
 import org.apache.hc.core5.io.CloseMode;
 import org.apache.hc.core5.net.NamedEndpoint;
@@ -93,7 +97,7 @@ import org.junit.jupiter.params.provider.ArgumentsProvider;
 import org.junit.jupiter.params.provider.ArgumentsSource;
 import org.junit.jupiter.params.provider.ValueSource;
 
-public class TLSIntegrationTest {
+class TLSIntegrationTest {
 
     private static final Timeout TIMEOUT = Timeout.ofSeconds(30);
 
@@ -133,7 +137,7 @@ public class TLSIntegrationTest {
 
     HttpAsyncServer createServer(final TlsStrategy tlsStrategy) {
         return AsyncServerBootstrap.bootstrap()
-                .setLookupRegistry(new UriPatternMatcher<>())
+                .setCanonicalHostName("localhost")
                 .setIOReactorConfig(
                         IOReactorConfig.custom()
                                 .setSoTimeout(TIMEOUT)
@@ -144,7 +148,11 @@ public class TLSIntegrationTest {
                 .setIOSessionDecorator(LoggingIOSessionDecorator.INSTANCE)
                 .setExceptionCallback(LoggingExceptionCallback.INSTANCE)
                 .setIOSessionListener(LoggingIOSessionListener.INSTANCE)
-                .register("*", () -> new EchoHandler(2048))
+                .setIOReactorMetricsListener(LoggingReactorMetricsListener.INSTANCE)
+                .setRequestRouter(RequestRouter.<Supplier<AsyncServerExchangeHandler>>builder()
+                        .addRoute(RequestRouter.LOCAL_AUTHORITY, "*", () -> new EchoHandler(2048))
+                        .resolveAuthority(RequestRouter.LOCAL_AUTHORITY_RESOLVER)
+                        .build())
                 .create();
     }
 
@@ -159,11 +167,12 @@ public class TLSIntegrationTest {
                 .setIOSessionDecorator(LoggingIOSessionDecorator.INSTANCE)
                 .setExceptionCallback(LoggingExceptionCallback.INSTANCE)
                 .setIOSessionListener(LoggingIOSessionListener.INSTANCE)
+                .setIOReactorMetricsListener(LoggingReactorMetricsListener.INSTANCE)
                 .create();
     }
 
     Future<TlsDetails> executeTlsHandshake() throws Exception {
-        final Future<ListenerEndpoint> future = server.listen(new InetSocketAddress(0), URIScheme.HTTPS);
+        final Future<ListenerEndpoint> future = server.listen(new InetSocketAddress(InetAddress.getLoopbackAddress(), 0), URIScheme.HTTPS);
         final ListenerEndpoint listener = future.get(TIMEOUT.getDuration(), TIMEOUT.getTimeUnit());
         final InetSocketAddress address = (InetSocketAddress) listener.getAddress();
 
@@ -180,14 +189,7 @@ public class TLSIntegrationTest {
                         try {
                             ((TlsUpgradeCapable) clientEndpoint).tlsUpgrade(
                                     target,
-                                    new FutureContribution<ProtocolIOSession>(tlsFuture) {
-
-                                        @Override
-                                        public void completed(final ProtocolIOSession protocolIOSession) {
-                                            tlsFuture.completed(protocolIOSession.getTlsDetails());
-                                        }
-
-                                    });
+                                    new CompletingFutureContribution<>(tlsFuture, ProtocolIOSession::getTlsDetails));
                         } catch (final Exception ex) {
                             tlsFuture.failed(ex);
                         }
@@ -199,7 +201,7 @@ public class TLSIntegrationTest {
 
     @ParameterizedTest(name = "TLS protocol {0}")
     @ArgumentsSource(SupportedTLSProtocolProvider.class)
-    public void testTLSSuccess(final TLS tlsProtocol) throws Exception {
+    void testTLSSuccess(final TLS tlsProtocol) throws Exception {
         final TlsStrategy serverTlsStrategy = new TestTlsStrategy(
                 SSLTestContexts.createServerSSLContext(),
                 (endpoint, sslEngine) -> sslEngine.setEnabledProtocols(new String[]{tlsProtocol.getId()}),
@@ -225,7 +227,7 @@ public class TLSIntegrationTest {
     }
 
     @Test
-    public void testTLSTrustFailure() throws Exception {
+    void testTLSTrustFailure() throws Exception {
         final TlsStrategy serverTlsStrategy = new BasicServerTlsStrategy(SSLTestContexts.createServerSSLContext());
         server = createServer(serverTlsStrategy);
         server.start();
@@ -243,7 +245,7 @@ public class TLSIntegrationTest {
     }
 
     @Test
-    public void testTLSClientAuthFailure() throws Exception {
+    void testTLSClientAuthFailure() throws Exception {
         final TlsStrategy serverTlsStrategy = new BasicServerTlsStrategy(
                 SSLTestContexts.createServerSSLContext(),
                 (endpoint, sslEngine) -> sslEngine.setNeedClientAuth(true),
@@ -255,7 +257,7 @@ public class TLSIntegrationTest {
         client = createClient(clientTlsStrategy);
         client.start();
 
-        final Future<ListenerEndpoint> future = server.listen(new InetSocketAddress(0), URIScheme.HTTPS);
+        final Future<ListenerEndpoint> future = server.listen(new InetSocketAddress(InetAddress.getLoopbackAddress(), 0), URIScheme.HTTPS);
         final ListenerEndpoint listener = future.get(TIMEOUT.getDuration(), TIMEOUT.getTimeUnit());
         final InetSocketAddress address = (InetSocketAddress) listener.getAddress();
 
@@ -273,7 +275,7 @@ public class TLSIntegrationTest {
     }
 
     @Test
-    public void testSSLDisabledByDefault() throws Exception {
+    void testSSLDisabledByDefault() throws Exception {
         final TlsStrategy serverTlsStrategy = new TestTlsStrategy(
                 SSLTestContexts.createServerSSLContext(),
                 (endpoint, sslEngine) -> sslEngine.setEnabledProtocols(new String[]{"SSLv3"}),
@@ -309,7 +311,7 @@ public class TLSIntegrationTest {
             "TLS_KRB5_EXPORT_WITH_RC4_40_SHA",
             "SSL_RSA_EXPORT_WITH_RC2_CBC_40_MD5"
     })
-    public void testWeakCipherDisabledByDefault(final String cipher) throws Exception {
+    void testWeakCipherDisabledByDefault(final String cipher) throws Exception {
         final TlsStrategy serverTlsStrategy = new TestTlsStrategy(
                 SSLTestContexts.createServerSSLContext(),
                 (endpoint, sslEngine) -> sslEngine.setEnabledCipherSuites(new String[]{cipher}),
@@ -328,7 +330,7 @@ public class TLSIntegrationTest {
     }
 
     @Test
-    public void testTLSVersionMismatch() throws Exception {
+    void testTLSVersionMismatch() throws Exception {
         final TlsStrategy serverTlsStrategy = new TestTlsStrategy(
                 SSLTestContexts.createServerSSLContext(),
                 (endpoint, sslEngine) -> {
@@ -357,6 +359,40 @@ public class TLSIntegrationTest {
         Assertions.assertInstanceOf(IOException.class, cause);
     }
 
+    @Test
+    void testHostNameVerification() throws Exception {
+        server = createServer(new BasicClientTlsStrategy(SSLTestContexts.createServerSSLContext()));
+        server.start();
+
+        client = createClient(new BasicClientTlsStrategy(SSLTestContexts.createClientSSLContext()));
+        client.start();
+
+        final Future<ListenerEndpoint> future = server.listen(new InetSocketAddress(InetAddress.getLoopbackAddress(), 0), URIScheme.HTTPS);
+        final ListenerEndpoint listener = future.get(TIMEOUT.getDuration(), TIMEOUT.getTimeUnit());
+        final InetSocketAddress address = (InetSocketAddress) listener.getAddress();
+
+        final HttpHost target1 = new HttpHost(URIScheme.HTTPS.id, InetAddress.getLoopbackAddress(), "localhost", address.getPort());
+        final Future<Message<HttpResponse, String>> resultFuture1 = client.execute(
+                target1,
+                new BasicRequestProducer(Method.POST, target1, "/stuff",
+                        new StringAsyncEntityProducer("some stuff", ContentType.TEXT_PLAIN)),
+                new BasicResponseConsumer<>(new StringAsyncEntityConsumer()), TIMEOUT, null);
+        final Message<HttpResponse, String> message1 = resultFuture1.get(TIMEOUT.getDuration(), TIMEOUT.getTimeUnit());
+        Assertions.assertNotNull(message1);
+        Assertions.assertEquals(200, message1.getHead().getCode());
+
+        final HttpHost target2 = new HttpHost(URIScheme.HTTPS.id, InetAddress.getLoopbackAddress(), "some-other-host", address.getPort());
+        final Future<Message<HttpResponse, String>> resultFuture2 = client.execute(
+                target2,
+                new BasicRequestProducer(Method.POST, target2, "/stuff",
+                        new StringAsyncEntityProducer("some stuff", ContentType.TEXT_PLAIN)),
+                new BasicResponseConsumer<>(new StringAsyncEntityConsumer()), TIMEOUT, null);
+        final ExecutionException exception = Assertions.assertThrows(ExecutionException.class, () ->
+                resultFuture2.get(TIMEOUT.getDuration(), TIMEOUT.getTimeUnit()));
+        final Throwable cause = exception.getCause();
+        Assertions.assertInstanceOf(SSLHandshakeException.class, cause);
+    }
+
     static class SupportedTLSProtocolProvider implements ArgumentsProvider {
 
         int javaVere = ReflectionUtils.determineJRELevel();
